

// // Offset between JD and MJD
// pub const MJD2JD: f64 = 2400000.5;

// Speed of light, MKS, (exact value)
pub const C: f64 = 2.99792458e8;  

// // Gravitational constant, MKS
// pub const G: f64 = 6.673e-11;

// Planck's constant, MKS
pub const H: f64 = 6.6262e-34;

// Boltzmann's constant, MKS
pub const K: f64 = 1.3806e-23;  

// // Charge on the electron (magnitude, C)
// pub const E: f64 = 1.602176565e-19;

// // Mass of the electron, kg
// pub const ME: f64 = 9.10956e-31; 

// // Mass of the proton, kg
// pub const MP: f64 = 1.67e-27;  

// // Stefan-Boltzmann constant, MKS
// pub const SIGMA: f64 = 5.66956e-8;

// // Thomson cross-section, MKS
// pub const SIGMAT: f64 = 6.65e-29;

// // Astronomical unit, metres
// pub const AU: f64 = 1.49597870691e11;

// // Solar luminosity, Watts
// pub const LSUN: f64 = 3.826e26;

// // Solar mass, kg    
// pub const MSUN: f64 = 1.989e30;

// // Gravitational parameter of the Sun, SI (m^3 s^-2)
// pub const GMSUN: f64 = 1.32712442099e20;

// // Gravitational parameter of the Sun, AU^3 YR^-2
// pub const GMSUNA: f64 = 39.476927033270655;

// // Gauss' gravitational constant sqrt(G*MSUN), AU^(3/2) day^-1
// pub const KGAUSS: f64 = 0.01720209895;     

// // G*MSUN, AU^3 day^-2 (Gauss**2)
// pub const GMGAUSS: f64 = KGAUSS*KGAUSS;     

// // Absolute visual magnitude of the Sun
// pub const MVSUN: f64 = 4.75; 

// // Parsec, metres
// pub const PC: f64 = 3.085678e16;

// // Solar radius, metres
// pub const RSUN: f64 = 6.9599e8;

// // Effective temperature of the Sun, Kelvin
// pub const TSUN: f64 = 5700.;  

// Number of seconds in a day
pub const DAY: f64 = 86400.;

// // Length of Julian year in seconds 
// pub const YEAR: f64 = 365.25*DAY;

// // Integer number of seconds in a day
// pub const IDAY: u32  = 86400;

// // Number of seconds in an hour
// pub const HOUR: f64  = 3600.;     

// // Number of seconds in a minute
// pub const MINUTE: f64  = 60.;     

// // Pi
// pub const PI: f64 = 3.14159265358979323846264;  

// // 2*Pi
// pub const TWOPI: f64 = 2.*PI; 

// Ratio FWHM/sigma for a gaussian
pub const EFAC: f64 = 2.3548200450309493;

// // Wavelength of Halpha, Angstroms
// pub const HALPHA: f64 = 6562.76; 

// // Wavelength of Hbeta, Angstroms
// pub const HBETA: f64 = 4861.327; 

// // Wavelength of Hgamma, Angstroms
// pub const HGAMMA: f64 = 4340.465; 

// // Wavelength of Hgamma, Angstroms
// pub const HDELTA: f64 = 4340.465;